/*
 * Copyright (C) 2001, 2002, NRFS Project Team
 */

#ifndef __BUFFER_INCLUDE__
#define __BUFFER_INCLUDE__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>

#define U32_LENGTH		4

#define SERVER_MAX		32
#define BUFFER_SIZE_U32		1024
#define BUFFER_SIZE_BYTE	( BUFFER_SIZE_U32 << 2 )

typedef struct
{
	char *	buffer;
	int	offset;
	int	length;
} buffer_t;
typedef buffer_t BUFFER;

typedef struct
{
        unsigned long           xid;
        unsigned long           cip;
        uid_t                   uid;
        gid_t                   gid;

        int                     serv_ok;
        int                     serv_ln;
        int                     serv_cn;
        struct
        {
                unsigned long   serv_ip;
                char *          serv_ex;
                struct sockaddr_in
                                serv_sk;
                int             serv_fd;
                int             serv_at;
                unsigned long   type;
                struct stat     stat;
                unsigned long   csum;
                unsigned long   mark;
        } serv_rp[ SERVER_MAX ];

        long                    targ_ch;
        char *                  targ_pt;
        unsigned long           targ_of;
        unsigned long           targ_ln;

        char                    sbuff[ BUFFER_SIZE_BYTE ];
        buffer_t                sendbuff;
        char                    rbuff[ BUFFER_SIZE_BYTE ];
        buffer_t                recvbuff;
} recovery_t;

typedef struct
{
	unsigned long	cip;
	unsigned long	serv_ip;
	char *		serv_ex;

	unsigned long	targ_ct;
	unsigned long	targ_ch;
	unsigned long	targ_of;
	unsigned long	targ_ln;
	char *		targ_pt;
	unsigned long	targ_type;
	struct stat	targ_st;

        char                    pbuff[ BUFFER_SIZE_BYTE ];
        buffer_t                pathbuff;
        char                    wbuff[ BUFFER_SIZE_BYTE ];
        buffer_t                workbuff;
        char                    sbuff[ BUFFER_SIZE_BYTE ];
        buffer_t                sendbuff;
        char                    rbuff[ BUFFER_SIZE_BYTE ];
        buffer_t                recvbuff;
} check_t;

typedef struct
{
	unsigned long	targ_ct;
	char *		targ_pt;
	unsigned long	targ_type;
	struct stat	targ_st;

        char                    pbuff[ BUFFER_SIZE_BYTE ];
        buffer_t                pathbuff;
        char                    sbuff[ BUFFER_SIZE_BYTE ];
        buffer_t                sendbuff;
        char                    rbuff[ BUFFER_SIZE_BYTE ];
        buffer_t                recvbuff;
} repair_t;


struct pool_buffer
{
	size_t			total_size;
	size_t			free_size;
	struct pool_buffer *	next;
	unsigned char *		offset;
};
typedef struct pool_buffer  pool_t;

#define POOL_HEAD_SIZE          ( sizeof( pool_t ) )
#define POOL_BUFFER_SIZE(s)     ( ( s + POOL_HEAD_SIZE + 4095 ) & (~0xFFF))

/*
 * pool buffer functions
 */
static inline pool_t *
__create_pool( size_t size )
{
	pool_t *pool;

	if ( ( pool = (pool_t *)malloc( size ) ) )
	{
		pool->total_size = size - POOL_HEAD_SIZE;
		pool->free_size  = pool->total_size;
		pool->offset = (char *)&pool[1];
		pool->next = NULL;
	}
	return pool;
}

static inline void *
__alloc_pool( pool_t * pool, size_t size )
{
	void *  buff = NULL;

	size = (( size + 3 )&(~3));
	if ( size <= pool->free_size )
	{
		buff = pool->offset;
		pool->free_size -= size;
		pool->offset    += size;
	}
	return buff;
}

static inline pool_t *
create_pool( size_t size )
{
	return __create_pool( POOL_BUFFER_SIZE(size) );
}

static inline void
delete_pool( pool_t * pool )
{
	if ( pool )
	{
		delete_pool( pool->next );
		pool->next = NULL;
		free( pool );
	}
}

static inline void
clear_pool( pool_t * pool )
{
	if ( pool )
	{
		delete_pool( pool->next );
		pool->next = NULL;
		pool->free_size  = pool->total_size;
		pool->offset = (char *)&pool[1];
	}
}

static inline void *
pool_alloc( pool_t * pool, size_t size )
{
	void *  buff = NULL;

	if ( pool )
	{
		while( !( buff = __alloc_pool( pool, size ) ) && pool->next )
			pool = pool->next;
		if ( !buff )
		{
			pool->next = __create_pool( POOL_BUFFER_SIZE(size) );
			buff = __alloc_pool( pool->next, size );
		}
	}
	return buff;
}


/*
 * en/decode functions
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static inline unsigned long *
get_u32ptr( buffer_t * buf )
{
	return ( unsigned long * )&( buf->buffer[ buf->offset ] );
}

static inline int
quad_u32len( int len )
{
	return ( ( len + 3 ) & (~3) );
}

static inline int
to_u32len( int len )
{
	return ( ( len + 3 ) >> 2 );
}

static inline int
to_bytelen( int len )
{
	return ( len << 2 );
}

static inline int
check_length( buffer_t * buf, int n )
{
	return ( ( buf->length - buf->offset ) >= n );
}

static inline int
check_u32length( buffer_t * buf, int n )
{
	return ( ( buf->length - buf->offset ) >= to_bytelen( n ) );
}

static inline unsigned long
get_hlong( buffer_t * buf )
{
	unsigned long val = *get_u32ptr( buf );
	buf->offset += U32_LENGTH;
	return ntohl( val );
}

static inline unsigned long
get_nlong( buffer_t * buf )
{
	unsigned long val = *get_u32ptr( buf );
	buf->offset += U32_LENGTH;
	return val;
}

static inline char *
get_hstr( buffer_t * buf )
{
	if ( check_u32length( buf, 1 ) )
	{
		char *	str = &( buf->buffer[ buf->offset ] );
		int	len = get_hlong( buf );

		if ( check_length( buf, len ) )
		{
			memmove( str, str + U32_LENGTH, len );
			str[ len ] = '\0';
			buf->offset += quad_u32len( len );
			return str;
		}
		buf->offset = buf->length;
	}
	return NULL;
}

static inline char *
get_nstr( buffer_t * buf )
{
	if ( check_u32length( buf, 1 ) )
	{
		char *	str = &( buf->buffer[ buf->offset ] );
		int	len = get_hlong( buf );

		if ( check_length( buf, len ) )
		{
			buf->offset += quad_u32len( len );
			return str;
		}
		buf->offset = buf->length;
	}
	return NULL;
}

static inline unsigned long *
set_u32ptr( buffer_t * buf )
{
	return ( unsigned long * )&( buf->buffer[ buf->length ] );
}

static inline int
free_length( buffer_t * buf, int n )
{
	return ( ( BUFFER_SIZE_BYTE - buf->length ) >= n );
}

static inline int
free_u32length( buffer_t * buf, int n )
{
	return ( ( BUFFER_SIZE_BYTE  - buf->length ) >= to_bytelen( n ) );
}

static inline void
set_hlong( buffer_t * buf, unsigned long hl )
{
	*set_u32ptr( buf ) = htonl( hl );
	buf->length += U32_LENGTH;
}

static inline void
set_nlong( buffer_t * buf, unsigned long nl )
{
	*set_u32ptr( buf ) = nl;
	buf->length += U32_LENGTH;
}

static inline int
set_hstr( buffer_t * buf, char *hstr )
{
	if ( hstr )
	{
		int	slen = strlen( hstr ), ulen = to_u32len( slen );

		if ( free_length( buf, U32_LENGTH + quad_u32len( slen ) ) )
		{
			set_hlong( buf, slen );
			memcpy( &( buf->buffer[ buf->length ] ), hstr, slen );
			buf->length += quad_u32len( slen );
			return 1;
		}
	}
	return 0;
}

static inline int
set_nstr( buffer_t * buf, unsigned long * nstr )
{
	if ( nstr )
	{
		int	slen = ntohl( *nstr++ );

		if ( free_length( buf, U32_LENGTH + quad_u32len( slen ) ) )
		{
			set_hlong( buf, slen );
			memcpy( &( buf->buffer[ buf->length ] ), nstr, slen );
			buf->length += quad_u32len( slen );
			return 1;
		}
	}
	return 0;
}
#endif
