/*
 * Copyright (C) 2001, 2002, NRFS Project Team
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <termios.h>
#include <time.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>

#include "recovery.h"

static inline int
tcp_canread_one( int fd )
{
	fd_set		fds;
	struct timeval	tout;

	FD_ZERO( &fds );
	FD_SET( fd, &fds );
#if 1
	tout.tv_sec = 0;
	tout.tv_usec  = TCP_TIMEOUT_ONE;
#else
	tout.tv_sec  = TCP_TIMEOUT_ONE;
	tout.tv_usec = 0;
#endif

	return select( FD_SETSIZE, &fds,
			( fd_set * )NULL, ( fd_set * )NULL, &tout );
}

int
tcp_canread( int fd )
{
	int	state = 0, count = TCP_TIMEOUT_TRY;

	while ( count -- )
		if ( ( state = tcp_canread_one( fd ) ) )
			break;

	return state;
}

static inline int
tcp_canwrite_one( int fd )
{
	fd_set		fds;
	struct timeval	tout;

	FD_ZERO( &fds );
	FD_SET( fd, &fds );
#if 1
	tout.tv_sec = 0;
	tout.tv_usec  = TCP_TIMEOUT_ONE;
#else
	tout.tv_sec  = TCP_TIMEOUT_ONE;
	tout.tv_usec = 0;
#endif

	return select( FD_SETSIZE, ( fd_set * )NULL,
			&fds, ( fd_set * )NULL, &tout );
}

int
tcp_canwrite( int fd )
{
	int	state = 0, count = TCP_TIMEOUT_TRY;

	while ( count -- )
		if ( ( state = tcp_canwrite_one( fd ) ) )
			break;

	return state;
}

int
tcp_bind( struct sockaddr_in * sock )
{
	int	state = -1;

	if ( sock && ( state = socket( PF_INET, SOCK_STREAM, 0 ) ) != -1 )
	{
		int	fd = state, slen = sizeof( struct sockaddr_in );
	
		if ( ( state = bind( fd, sock, slen ) ) != -1 &&
				( state = listen( fd, 5 ) ) != -1 )
			state = fd;
		else
			close( fd );
	}
	return state;
}

int
tcp_accept( int fd, struct sockaddr_in * sock )
{
	int	state = -1;

	if ( fd != -1 && sock && ( state = tcp_canread_one( fd ) ) > 0 )
	{
		int	slen = sizeof( struct sockaddr_in );
#if 1
		state = accept( fd, sock, &slen );
#else
		if ( ( state = accept( fd, sock, &slen ) ) != -1 )
		{
			time_t t = time( NULL );
			fprintf( stdFp, "accept client %s : %s",
					inet_ntoa( sock->sin_addr.s_addr ),
					ctime( &t ) );
		}
#endif
	}
	else
	{
		state = -1;
	}
	return state;
}

int
tcp_connect( struct sockaddr_in * sock )
{
	int	state = -1;

	if ( sock && ( state = socket( PF_INET, SOCK_STREAM, 0 ) ) != -1 )
	{
		int	fd = state, slen = sizeof( struct sockaddr_in );

		if ( ( state = tcp_canwrite( fd ) ) > 0 &&
			!( state = connect( fd, sock, slen ) ) )
		{
#if 0
			time_t t = time( NULL );
			fprintf( stdFp, "connect server %s : %s",
					inet_ntoa( sock->sin_addr.s_addr ),
					ctime( &t ) );
#endif
			state = fd;
		}
		else
		{
			close( fd );
			state = -1;
		}
	}

	return state;
}

int
tcp_read_one( int fd, unsigned char * buf, int len )
{
	int	state = tcp_canread_one( fd );

	if ( state > 0 )
		return read( fd, buf, len );
	return state;
}

int
tcp_read( int fd, unsigned char * buf, int len )
{
	if ( fd != -1 && buf )
	{
		int	rc = 0, c, count = TCP_TIMEOUT_TRY;

		while( count-- )
		{
			c = tcp_read_one( fd, buf + rc, len - rc );
			if ( c < 0 || ( rc += c ) == len )
				break;
		}
		return ( ( c < 0 ) ? c : rc );
	}
	return fd;
}

int
tcp_write_one( int fd, unsigned char * buf, int len )
{
	int	state = tcp_canwrite_one( fd );

	if ( state > 0 )
		return write( fd, buf, len );
	return state;
}

int
tcp_write( int fd, unsigned char * buf, int len )
{
	if ( fd != -1 && buf )
	{
		int	wc = 0, c, count = TCP_TIMEOUT_TRY;

		while( count -- )
		{
			c = tcp_write_one( fd, buf + wc, len - wc );
			if ( c < 0 || ( wc += c ) == len )
				break;
		}
		return ( ( c < 0 ) ? c : wc );
	}
	return fd;
}

int
receive_buffer( int fd, buffer_t * buff )
{
	if ( fd != -1 && buff )
		return ( buff->length = tcp_read( fd,
			&( buff->buffer[ buff->offset ] ), buff->length ) );
	return -1;
}

int
send_buffer( int fd, buffer_t * buff )
{
	if ( fd != -1 && buff )
		return ( buff->length = tcp_write( fd,
			&( buff->buffer[ buff->offset ] ), buff->length ) );
	return -1;
}

int
send_status( int fd, unsigned long status )
{
	int     state = -1;
	unsigned long buff = htonl( status );

	if ( tcp_write( fd, ( unsigned char * )&buff,
				U32_LENGTH ) == U32_LENGTH )
		state = 0;
	return  state;
}

int
receive_status( int fd, unsigned long * status )
{
	int     state = -1;
	unsigned long buff;

	if ( status && tcp_read( fd, ( unsigned char * )&buff,
				U32_LENGTH ) == U32_LENGTH )
	{
		*status = ntohl( buff );
		state = 0;
	}
	return state;
}

